# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

from .base import *
from .diluter import dilute
from .popDub import calculatePopulationDoublings
from .repeatSearch import searchSequenceForRepeats
from .Kmers import runOligoFreqAnalysis
from .complexity import *
from .MMBSearchTK import *
from .deepSeqInsH import *
from .variantTK import *
from .paralleltools import *


# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

## Contents:
## qualitySNPFilter
## filterByAD
## findDominantAF
## findType
## findZygosity
## filterFromClones
## drawSNPMap
## findGenomeLength
## generateRandomLoci
## findChrInLinearGenome
## findSpectraStrandwise
## renameChrToRoman
## draw_random_SNPMap
## drawCombinedSNPMap

import os
import random
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from natsort import index_natsorted
import seaborn as sns
from numpy.core.numeric import NaN


def qualitySNPFilter(frames_list):

  for i in range(len(frames_list)):
    frames_list[i] = frames_list[i][frames_list[i].TYPE =='SNP']
    frames_list[i] = frames_list[i].rename(columns=lambda c: 'AF' if 'AF' in c else c)
    frames_list[i] = frames_list[i].rename(columns=lambda c: 'AD' if 'AD' in c else c)
    
    #this line may be problematic for a good result and if not needed comment out
    frames_list[i]['AF'] = frames_list[i].apply(lambda row: findDominantAF(row), axis=1)

    frames_list[i] = frames_list[i].astype({'AF': float})
    frames_list[i] = frames_list[i][frames_list[i].AF >= 0.35]
    #also filter out 1.5n variants?

    frames_list[i]['EVAL_AD'] = frames_list[i].apply(lambda row: filterByAD(row), axis=1)
    frames_list[i] = frames_list[i][frames_list[i].EVAL_AD == 'True']
    frames_list[i]['SPECTRA'] = frames_list[i].apply(lambda row: findType(row.REF, row.ALT), axis=1)
  return frames_list

def filterByAD(row):
  reads=row['AD'].split(",")
  if int(reads[1]) < 5:
    return "False"
  if int(reads[0])+int(reads[1]) > 9:
    return "True"
  else:
    return "False"

def findDominantAF(row):
  reads=row['AF'].split(",")
  return(max(reads))

def findType(colA, colB):
  if ((colA == 'C') & (colB == 'T') | (colA == 'G') & (colB == 'A')):
    return('C_to_T')
  if ((colA == 'C') & (colB == 'A') | (colA == 'G') & (colB == 'T')):
    return('C_to_A')
  if ((colA == 'C') & (colB == 'G') | (colA == 'G') & (colB == 'C')):
    return('C_to_G')
  if ((colA == 'T') & (colB == 'C') | (colA == 'A') & (colB == 'G')):
    return('T_to_C')
  if ((colA == 'T') & (colB == 'G') | (colA == 'A') & (colB == 'C')):
    return('T_to_G')
  if ((colA == 'T') & (colB == 'A') | (colA == 'A') & (colB == 'T')):
    return('T_to_A')

def findZygosity(row):
  if (row["AF"] >= 0.85):
    return('Homozygous')
  else:
    return('Heterozygous')

def filterFromClones(frames_list):
  #this will subtract all common rows between normal/control and experimental
  #sample. I might need to revist to filter variants by position only not all columns
  subtracted_df_list = []
  for df in frames_list:
    temp_df = df.drop(columns=['AD','AF'])
    len_original=len(temp_df)
    for df_2 in frames_list:
      if not df.equals(df_2):
        df_2=df_2.drop(columns=['AD','AF'])
        temp_df = pd.merge(temp_df,df_2, indicator=True, how='outer').query('_merge=="left_only"').drop('_merge', axis=1)

    len_final=len(temp_df)
    print(f"removed {len_original-len_final} rows from df")
    subtracted_df_list.append(temp_df)
  return subtracted_df_list

def drawSNPMap(pd_df, df_chr_lengths, chr_starts_df, title, sample_names, saveMap=True):

  pd_df = pd_df.append(chr_starts_df, ignore_index=True)

  fig, ax = plt.subplots()

  df_chr_lengths.plot(kind='barh',legend=False, ax=ax, x="chromosome", y="end_position", fontsize=15, figsize=(20,6), edgecolor="black", linewidth=1, color="beige", zorder=0, title=title, label="Position")
  
  pd_df = pd_df.sort_values(by="CHROM", key=lambda x: np.argsort(index_natsorted(pd_df["CHROM"])))
  
  #To rename legend elements, change plot markers/colors, modify here
  label_dict = { "G": "G->N", "C": "C->N", "A": "A->N", "T": "T->N", "REF": "Reference Allele"}
  markers = {"Homozygous": "x", "Heterozygous": "|"}
  palette = {"G": "green", "C": "red", "A": "gray", "T": "gray"}

  sns.scatterplot(ax=ax,data=pd_df, x="POS", y="CHROM", hue="REF", palette=palette, style="ZYGOSITY", markers=markers, s=120, alpha=0.7,zorder=1, linewidth=1.5)
  
  ax.set_xlabel("POSITION")
  ax.set_ylabel("CHROMOSOME")

  legend_texts = plt.legend().get_texts()
  for i in range(len(legend_texts)):
    label_str = legend_texts[i].get_text()
    if label_str in label_dict:
      new_label = label_dict.get(label_str)

      legend_texts[i].set_text(new_label)

  if saveMap:
    plt.savefig('figures/'+title+'.png', transparent=False)

def findGenomeLength(chromosomes):

  genome_length = 0
  for chr in chromosomes:
    genome_length+=chr[1]

  return(genome_length)

def generateRandomLoci(n, chromosomes):

  genome_length=findGenomeLength(chromosomes)
  return(random.sample(range(0, genome_length), n))

def findChrInLinearGenome(chromosomes, random_loci):

  output_list=[]
  for locus in random_loci:
    linear_genome_slider=0
    linear_genome_slider_previous=0
    for chromosome in chromosomes:
      linear_genome_slider+=chromosome[1]

      if locus < linear_genome_slider:
        #print(f'slider:{linear_genome_slider}, chr:{chromosome[0]}, locus:{locus}')
        output_list.append([chromosome[0], locus-linear_genome_slider_previous])
        linear_genome_slider_previous+=chromosome[1]
        break
      linear_genome_slider_previous+=chromosome[1]

  return(output_list)

def findSpectraStrandwise(colA, colB):
  if ((colA == 'C') & (colB == 'T')):
    return('C_to_T')
  elif ((colA == 'C') & (colB == 'A')):
    return('C_to_A')
  elif ((colA == 'C') & (colB == 'G')):
    return('C_to_G')
  elif ((colA == 'T') & (colB == 'C')):
    return('T_to_C')
  elif ((colA == 'T') & (colB == 'G')):
    return('T_to_G')
  elif ((colA == 'T') & (colB == 'A')):
    return('T_to_A')

  elif ((colA == 'G') & (colB == 'A')):
    return('G_to_A')
  elif ((colA == 'G') & (colB == 'T')):
    return('G_to_T')
  elif ((colA == 'G') & (colB == 'C')):
    return('G_to_C')
  elif ((colA == 'A') & (colB == 'G')):
    return('A_to_G')
  elif ((colA == 'A') & (colB == 'C')):
    return('A_to_C')
  elif ((colA == 'A') & (colB == 'T')):
    return('A_to_T')

def renameChrToRoman(df):
  roman_names={"chr1": "I",
               "chr2": "II",
               "chr3": "III",
               "chr4": "IV",
               "chr5": "V",
               "chr6": "VI",
               "chr7": "VII",
               "chr8": "VIII",
               "chr9": "IX",
               "chr10": "X",
               "chr11": "XI",
               "chr12": "XII",
               "chr13": "XIII",
               "chr14": "XIV",
               "chr15": "XV",
               "chr16": "XVI"}
  df.replace({"chromosome": roman_names}, inplace=True)
  return(df)

def draw_random_SNPMap(pd_df, df_chr_lengths, chr_starts_df, title, saveMap=True):

  #To rename legend elements, change plot markers/colors, modify here

  label_dict = {"cen": 'Centromere'}
  markers = {"random": "$|$", "cen": "o"}
  palette = {"random": "black", "cen": "white"}

  
  pd_df = pd_df.append(chr_starts_df, ignore_index=True)
  pd_df = pd_df.sort_values(by="CHROM", key=lambda x: np.argsort(index_natsorted(pd_df["CHROM"])))
  renameChrToRoman(pd_df)


  fig, ax = plt.subplots()

  renameChrToRoman(df_chr_lengths)

  df_chr_lengths.plot(kind='barh',legend=False, ax=ax, x="chromosome", y="end_position", fontsize=40, figsize=(80,24), edgecolor="black", linewidth=1, color="lightgray", zorder=0, label="Position")
  sns.scatterplot(ax=ax, data=pd_df, x="POS", y="CHROM", style="REF", markers=markers, hue="REF", palette=palette, s=1500, alpha=1,zorder=1, linewidth=1, edgecolor="black", legend=False)

  ax.set_xlabel("POSITION", fontsize=40)
  ax.set_ylabel("CHROMOSOME", fontsize=40)
  fig.suptitle(title, fontsize=60)

  legend_texts = plt.legend().get_texts()
  for i in range(len(legend_texts)):
    label_str = legend_texts[i].get_text()
    if label_str in label_dict:
      new_label = label_dict.get(label_str)

      legend_texts[i].set_text(new_label)

  if saveMap:
    plt.savefig('figures/'+title+'.png', transparent=False, dpi=200)

def drawCombinedSNPMap(pd_df, df_chr_lengths, chr_starts_df, title, sample_names, saveMap=True):

  #To rename legend elements, change plot markers/colors, modify here
  label_dict = { "G": "G->N",
                 "C": "C->N",
                 "A": "A->N",
                 "T": "T->N",
                 "SPECTRA_STRANDWISE": "Reference Allele",
                 "cen": 'Centromere'}
                
  markers = {"C_to_T": "$|$",
             "C_to_A": "$|$",
             "C_to_G": "$|$",
             "G_to_A": "$|$",
             "G_to_C": "$|$",
             "G_to_T": "$|$",
             'T_to_A': "$|$",
             'T_to_C': "$|$",
             'A_to_C': "$|$",
             'A_to_G': "$|$",
             'A_to_T': "$|$",
             'T_to_G': "$|$",
             'cen': "o"}

  palette = {"C_to_T": "black",
             "C_to_A": "black",
             "C_to_G": "black",
             "G_to_A": "red",
             "G_to_C": "blue",
             "G_to_T": "olive",
             'T_to_A': "gray",
             'T_to_C': "gray",
             'A_to_C': "gray",
             'A_to_G': "gray",
             'A_to_T': "gray",
             'T_to_G': "gray",
             'cen': "white"}

  pd_df = pd_df.append(chr_starts_df, ignore_index=True)
  pd_df = pd_df.sort_values(by="CHROM", key=lambda x: np.argsort(index_natsorted(pd_df["CHROM"])))
  renameChrToRoman(pd_df)
  
  fig, ax = plt.subplots()

  renameChrToRoman(df_chr_lengths)
  df_chr_lengths.plot(kind='barh',legend=False, ax=ax, x="chromosome", y="end_position", fontsize=10, figsize=(20,6), edgecolor="black", linewidth=1, color="lightgray", zorder=0, label="Position")

  sns.scatterplot(ax=ax, data=pd_df[pd_df["REF"].isin(["C",'cen'])], x="POS", y="CHROM", hue="SPECTRA_STRANDWISE", palette=palette, style="SPECTRA_STRANDWISE", markers=markers, s=100, alpha=.9,zorder=1, linewidth=.05, edgecolor="black", legend = False)
  sns.scatterplot(ax=ax, data=pd_df[pd_df["REF"].isin(["G",'cen'])], x="POS", y="CHROM", hue="SPECTRA_STRANDWISE", palette=palette, style="SPECTRA_STRANDWISE", markers=markers, s=100, alpha=.9,zorder=2, linewidth=.15, edgecolor="black")

  ax.set_xlabel("POSITION", fontsize=10)
  ax.set_ylabel("CHROMOSOME", fontsize=10)
  fig.suptitle(title, fontsize=15)

  legend_texts = plt.legend().get_texts()
  for i in range(len(legend_texts)):
    label_str = legend_texts[i].get_text()
    if label_str in label_dict:
      new_label = label_dict.get(label_str)
      legend_texts[i].set_text(new_label)

  if saveMap:
    plt.savefig('figures/'+title+'.png', transparent=False, dpi=600)



# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

## Contents:
## dataFrameImport
## matchSequence
## findReadLength
## trimseq
## classify

import regex as re
import os
import pandas as pd

def dataFrameImport(directory,datatype = 'sam'):
  import pysam

  #This function imports all files inside of the specified DIR path into two lists; of panda DF and sample names
  frames_list_samples = []
  sample_names = []

  counter=1

  #import all files in the directory
  for filename in os.listdir(directory):
      if filename.endswith(f".{datatype}"):
          sample_names.append(filename[:8].strip())
          path = os.path.join(directory, filename)
          globals()[f"sample{counter}"] = pd.DataFrame(({'name': x.qname, 'seq': x.seq, 'qual': x.query_qualities} for x in pysam.Samfile(path).fetch()))
          print(f'dataframe {sample_names[-1]} has {len(globals()[f"sample{counter}"])} total rows')
          frames_list_samples.append(globals()[f"sample{counter}"])
          counter+=1
  print(f"found {len(frames_list_samples)} samples in {directory}")
  print(len(sample_names), sample_names)

  return(frames_list_samples, sample_names)

def matchSequence(row, sequence):
  #for searching for perfectly matched sequences.
  if (sequence in row):
    return(True)
  else:
    return(False)

def findReadLength(row):
  #returns a length of a read
  return(len(row))

def trimseq(row, f_seq, r_seq, consensus):
  #trims sequences to the 5' end of primers and finds reads that support non-deletions.
  #allows for 2 errors in non-deletions.

  if row.forward_primer == True:
    f_seq_index = row.seq.find(f_seq)
    if f_seq_index == -1:
      print("no forward primer found")
      row.forward_primer = False
    row.seq = row.seq[f_seq_index:]

  if row.reverse_primer == True:
    r_seq_index = row.seq.find(r_seq)
    if r_seq_index == -1:
      print("no reverse primer found")
      row.reverse_primer = False
    row.seq = row.seq[:r_seq_index+len(r_seq)]
  if (row.forward_primer == False) & (row.reverse_primer == False):
    row.seq = "no_primers"

  if re.search('('+row.seq+')'+'{e<=2}', consensus):
    row.seq = "no_excision"
  return(row)

def classify(row, consensus, classification):
  #this function classifies the sequence based on the provided consensus, allows for 2 errors.

  if row.classification != "other":
    return row

  if re.search('('+row.seq+')'+'{e<=2}', consensus):
  #if row.seq in consensus:
    row.classification = classification
  return row


# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

def calculatePopulationDoublings(initial_population: int|float, final_population_ml: int|float, final_culture_volume: int|float) -> None:
    """
    Calculates the number of population doublings required to reach a final population size.

    Args:
        initial_population (int|float): The initial population size.
        final_population_ml (int|float): The final population size in milliliters.
        final_culture_volume (int|float): The final culture volume in milliliters.

    Returns:
        None: This function does not return anything, but prints the number of population doublings required to reach the final population size.
    """
    final_population = final_population_ml * final_culture_volume
    doubling_cycle_size = final_population
    doublings = 0
    while doubling_cycle_size > initial_population:
        if doubling_cycle_size < initial_population*2:
            doublings = doublings + (doubling_cycle_size-initial_population)/initial_population
            break
        doubling_cycle_size = doubling_cycle_size/2.0
        doublings += 1
        print(f"Starting from the population size of {initial_population} to the final size of {final_population} cells,\\nthere were {doublings} population doublings")


# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

## Contents:
## findpairings
## findFrequencies
## plot
## runOligoFreqAnalysis

import pandas as pd
from .base import extractSeqFromFastaToList
from .base import validateSquence

def findpairings(sequence: str, pairing_length: int) -> list:
    '''
    Find all oligonucleotides of a given length in a DNA sequence.

    Args:
        sequence (str): the DNA sequence to search in.
        pairing_length (int): the length of the oligonucleotides to find.

    Returns:
        list: a list of all oligonucleotides of length 'pairing_length' found in 'sequence'.
    '''

    pairing_list = []
    for i in range(len(sequence)):
        next_chunk = sequence.upper()[i:i+pairing_length]
        if len(next_chunk) == pairing_length:
            pairing_list.append(next_chunk)
    return(pairing_list)

def findFrequencies(alist: list) -> dict:
    '''
    Count the frequency of each item in a list.

    Args:
        alist (list): the list to count the frequency of.

    Returns:
        dict: a dictionary where the keys are the items in the list and the values are their frequency.
    '''

    sequence_dict = {}

    for i in alist:
        if i in sequence_dict:
            sequence_dict[i]+=1
        else:
            sequence_dict[i] = 1
    return(sequence_dict)

def plot(oligo_dict: dict, title: str) -> None:
    '''
    Plot a histogram of the frequency of each item in a dictionary.

    Args:
        oligo_dict (dict): the dictionary to plot the frequency of.
        title (str): the title of the plot.

    Returns:
        None
    '''
    freq_list = list(oligo_dict.items())
    freq_list.sort(key=lambda x:x[1], reverse=True)
    df = pd.DataFrame(freq_list, columns=['oligo', 'frequency'])
    total_oligos = df.frequency.sum()
    df['frequency'] = df.frequency / total_oligos
    df.plot(kind='bar', x='oligo', figsize=(15,8), title=title+" Oligonucleotide Frequency", xlabel='Oligonucleotide', ylabel='Frequency (%)')

def runOligoFreqAnalysis(file_path: str) -> None:
    '''
    Run oligonucleotide frequency analysis on all sequences in a FASTA file.

    Args:
        file_path (str): the path to the FASTA file to analyze.

    Returns:
        None
    '''
    import matplotlib.pyplot as plt
    fa_seq_list = extractSeqFromFastaToList(file_path)

    for fa_seq in fa_seq_list:
        sequence = fa_seq[1]
        title = fa_seq[0]

        if not validateSquence(sequence):
            print(f'skipping "{title}", sequence validation failed')
            continue

        pairing_list_1 = findpairings(sequence,2)
        seqdeck_1 = findFrequencies(pairing_list_1)

        pairing_list_2 = findpairings(sequence,1)
        seqdeck_2 = findFrequencies(pairing_list_2)

        plot(seqdeck_1, title)
        plot(seqdeck_2, title)
        plt.show()


if __name__ == "__main__":
    import sys
    file_path = sys.argv[1]
    runOligoFreqAnalysis(file_path)




# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

## Contents:
## extractSeqFromFastaToList
## validateSquence
## compl
## rev_compl
## unique
## createChrList
## dataFrameImport

import logging

def extractSeqFromFastaToList(fasta_file_path: str) -> list:
    '''Extracts sequence information from a FASTA file and returns it as a nested list.

    Args:
        fasta_file_path (str): The path to the FASTA file.

    Returns:
        list: A nested list of the form [[title_0, sequence_0], [title_1, sequence_1], ...].
            Each element of the list is a list containing the title and sequence of a sequence in the FASTA file.
    '''
    fasta_file = open(fasta_file_path, 'r')
    contents = fasta_file.readlines()
    fasta_file.close()

    fasta_list=[]

    for i in contents:
        if '>' in i:
            fasta_list.append([i.strip('>').strip(),''])
        else:
            fasta_list[-1][1] = fasta_list[-1][1]+i.strip()
    logging.info(f"Extraction of sequence information from {fasta_file_path} finished.")
    return fasta_list

def validateSquence(sequence: str) -> bool:
    '''Checks if a DNA sequence contains only valid nucleotides.

    Args:
        sequence (str): A DNA sequence to be validated.

    Returns:
        bool: True if the sequence contains only valid nucleotides, False otherwise.
            If the sequence contains non-canonical nucleotides, a log message is generated.
            If the sequence contains 'N's, a warning log message is generated.
            The log messages are written to the default logger.
    '''
    bases = "ATGCatgcN"
    for i in sequence:
        if i not in bases:
            logging.info(f"Sequence doesn't contain cannonical nucleotides: {i}")
            return(False)
        if i == "N":
            logging.warning(f"Warning, sequence contains 'N's")
    return(True)

def compl(base: str) -> str:
    """
    Returns the complementary base for a given DNA base.

    Args:
        base (str): A single character representing a DNA base. Must be one of 'A', 'T', 'G', 'C', or '-'.

    Returns:
        str: The complementary base for the given DNA base. Returns '-' if the input is '-'.

    Raises:
        ValueError: If the input is not one of 'A', 'T', 'G', 'C', or '-'.
    """
    if base == "A":
        return('T')
    elif base == "T":
        return('A')
    elif base == "G":
        return('C')
    elif base == "C":
        return('G')
    elif base == "-":
        return('-')
    else:
        logging.error(f"Invalid base: {base}")
        raise ValueError(f"Invalid base: {base}")

def rev_compl(seq: str) -> str:
    """
    Returns the reverse complement of a given DNA sequence.

    Args:
        seq (str): A string representing a DNA sequence. Must only contain characters 'A', 'T', 'G', 'C', or '-'.

    Returns:
        str: The reverse complement of the given DNA sequence.

    Raises:
        ValueError: If the input sequence contains characters other than 'A', 'T', 'G', 'C', or '-' (Through compl())
    """
    new_seq = ""
    for base in seq:
        new_base = compl(base)
        new_seq = new_base + new_seq
    return(new_seq)

def unique(list1: list) -> list:
    """
    Returns a new list containing only the unique elements of the input list.

    Args:
        list1 (list): A list of elements.

    Returns:
        list: A new list containing only the unique elements of the input list.

    Examples:
        >>> unique([1, 2, 3, 2, 1])
        [1, 2, 3]

        >>> unique(['a', 'b', 'c', 'b', 'a'])
        ['a', 'b', 'c']
    """

    list_set = set(list1) 
    unique_list = (list(list_set))
    return unique_list

def createChrList(chrNum: int) -> list:
    """
    Creates a list of chromosome names.

    Args:
        chrNum (int): The number of chromosomes to include in the list.

    Returns:
        list: A list of chromosome names, where each name is a string of the form "chrX", where X is the chromosome number.

    Examples:
        >>> createChrList(3)
        ['chr1', 'chr2', 'chr3']

        >>> createChrList(5)
        ['chr1', 'chr2', 'chr3', 'chr4', 'chr5']
    """
    chrList = []
    for i in range(chrNum):
        chrList.append(f"chr{i+1}")
    return(chrList)

def dataFrameImport(directory: str) -> tuple:
    """
    Imports all TSV files in a directory as Pandas dataframes.

    Args:
        directory (str): The path to the directory containing the TSV files.

    Returns:
        tuple: A tuple containing two elements:
            - A list of Pandas dataframes, where each dataframe corresponds to a TSV file in the directory.
            - A list of sample names, where each name is a string representing the name of the corresponding TSV file.
    """
    import os
    import pandas as pd
    frames_list_samples = []
    sample_names = []

    counter=1
    for filename in os.listdir(directory):
            if filename.endswith(".tsv"):
                    sample_names.append(filename[:10].strip())
                    path = os.path.join(directory, filename)
                    globals()[f"sample{counter}"] = pd.read_csv(path, sep='\t')
                    print(f'dataframe {sample_names[-1]} has {len(globals()[f"sample{counter}"])} total rows')
                    frames_list_samples.append(globals()[f"sample{counter}"])
                    counter+=1
    logging.info(f"found {len(frames_list_samples)} samples in {directory}")
    logging.info(len(sample_names), sample_names)

    return(frames_list_samples, sample_names)

def pullGenomicContext(list_of_positions: list, fasta_file_path: str, context_flank: int = 5) -> list:
    """
    Extracts genomic context of a specified length around each position in a list of positions.

    Args:
        list_of_positions (list): A list of positions, where each position is a list of two elements: the chromosome name and the position.
        fasta_file_path (str): The path to the FASTA file containing the reference genome.
        context_flank (int, optional): The length of the context to extract on either side of each position. Defaults to 5.

    Returns:
        list: A list of genomic contexts, where each context is a list of four elements:
            - The sequence of bases to the left of the position.
            - The query base at the position.
            - The sequence of bases to the right of the position.
            - The chromosome name.
    """
    fasta_list = extractSeqFromFastaToList(fasta_file_path)
    context_list = []

    for i in fasta_list:
        logging.debug(f"extracting context from {i[0]}. It has {len(i[1])} bases.")
        sequence = i[1]
        for j in list_of_positions:

            #check if it is the same chromosome
            if j[0] != i[0]:
                continue
            #set j to be the position
            j = j[1]
            #find the position in the fasta list
            #remember that the fasta sequences start with 1, not 0, so we need to subtract 1 from the position

            try:
                query_base = sequence[j-1]

            except:
                logging.warning(f"position {j} not found in {i[0]}. Skipping...")
                continue

            #extract the context
            context_left = sequence[j-1-context_flank:j-1]
            context_right = sequence[j:j+context_flank]

            context_list.append([context_left, query_base, context_right, i[0]])

    return(context_list)

def drawGenomicContext(context_list: list, show: bool = False, **kwargs: dict) -> None:
    '''
    Draws a sequence logo of a list of genomic contexts.

    Args:
        context_list (list): A list of genomic contexts, where each context is a list of four elements:
            - The sequence of bases to the left of the position.
            - The query base at the position.
            - The sequence of bases to the right of the position.
            - The chromosome name.
        show (bool, optional): Whether to show the plot. Defaults to False.
        **kwargs (dict): Additional keyword arguments to pass to the function. Supported arguments include:
            - save_path (str): The path to save the plot to.

    Returns:
        None

    Raises:
        ImportError: If the required packages matplotlib.pyplot and pandas are not installed.
    '''
    
    import matplotlib.pyplot as plt
    import pandas as pd

    #create a dataframe from the context list
    df = pd.DataFrame(context_list, columns=['context_left', '0', 'context_right', 'chromosome'])

    #split the context into columns by base
    split_left = df['context_left'].apply(lambda x: pd.Series(list(x)))
    split_right = df['context_right'].apply(lambda x: pd.Series(list(x)))

    #rename the columns such that they are numbered. e.g. -3,-2,-1, 0, 1, 2, 3
    split_left.columns = [f"-{i}" for i in range(len(split_left.columns), 0, -1)]
    split_right.columns = [f"{i+1}" for i in range(len(split_right.columns))]

    #make a new dataframe with the context split into columns, query still in the middle
    df_context = pd.concat([split_left, df['0'], split_right], axis=1)

    df_context_freq = df_context.apply(pd.value_counts).fillna(0)              #get the frequencies of each base in each column
    df_context_freq = df_context_freq.div(df_context_freq.sum(axis=0), axis=1) #normalize the frequencies
    df_context_freq = df_context_freq.transpose()
    df_context_freq = df_context_freq*100                                      #set the frequencies to %
    
    #set color scheme for the bases so that they are consistent
    colors = {
                'A':'tab:green', 
                'T':'tab:red',
                'G':'tab:orange',
                'C':'tab:blue',
                'N':'tab:gray',
                }
    
    #plot the dataframe as a stacked barplot
    column_count = len(df_context_freq.columns) #count the number of columns
    df_context_freq.plot.bar(stacked=True, figsize=(column_count+3,4), color = list(colors.values()))
    plt.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))
    plt.ylabel('Frequency (%)')
    plt.xlabel('Relative Position')

    if 'save_path' in kwargs:
        plt.savefig(kwargs['save_path'], bbox_inches='tight', dpi=300)
        logging.info(f"saved plot to {kwargs['save_path']}")

    if show == True:
        plt.show()

    plt.close()



# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

## Contents:
## validateDNASquence
## imperfectHomologySearch
## findInvertedRepeat
## searchSequenceForRepeats
## saveToJSON

import regex as re

#from .base import validateSquence
from .base import rev_compl


def validateDNASquence(sequence: str) -> bool:
    '''
    Check if the given DNA sequence contains only valid nucleotides.

    Args:
        sequence (str): the DNA sequence to validate.

    Returns:
        bool: True if the sequence contains only valid nucleotides, False otherwise.
    '''

    bases = "ATGCatgc-"
    for i in sequence:
        if i not in bases:
            print("warning, sequence doesn't contain cannonical nucleotides")
            return(False)
    return(True)

def imperfectHomologySearch(sequence: str, query: str, min_homology: float = 0.8, fixed_errors: bool = False, inverted:bool = True) -> list:
    '''
    Search for a query sequence in a given DNA sequence, allowing for a certain number of mismatches/imperfect homology.

    Args:
        sequence (str): the DNA sequence to search in.
        query (str): the query sequence to search for.
        min_homology (float, optional): the minimum homology (similarity) between the query and the found sequence, as a fraction between 0 and 1. Defaults to 0.8.
        fixed_errors (bool, optional): if True, use a fixed number of errors instead of calculating it from min_homology. Defaults to False.
        inverted (bool, optional): if True, search for the reverse complement of the query sequence as well. Defaults to True.

    Returns:
        list: a list of query-match pairs, where each pair is a list containing the query sequence and a list of matching sequences.
    '''
    errors = 0
    if min_homology:
        errors = round(len(query)*(1-min_homology))
    if fixed_errors:
        errors = fixed_errors
    #print(f"searching with {errors} errors...")

    output_list = re.findall( '(' + query + '){e<=' + str(errors) + '}', sequence)

    if inverted == True:
        query=rev_compl(query)
    
    query_match_pairs=[query, output_list]
    if len(output_list) > 0:
        if len(output_list) > 1:
            print("This is unusual...")
        print(f'Found possible template(s) for {query}: {output_list}')
        return(query_match_pairs)
    else:
        return([])

def findInvertedRepeat(sequence: str, query_length: int = 4, min_spacer: int = 4, imperfect_homology: bool = False,
                        min_homology: float = 0.8, fixed_errors: bool = False, inverted: bool = True) -> list:
    '''
    Search for inverted repeats in a given DNA sequence.

    Args:
        sequence (str): the DNA sequence to search in.
        query_length (int, optional): the length of the query sequence to search for. Defaults to 4.
        min_spacer (int, optional): the minimum number of non-matching nucleotides between the two halves of the inverted repeat. Defaults to 4.
        imperfect_homology (bool, optional): if True, allow for a certain number of errors in the inverted repeat. Defaults to False.
        min_homology (float, optional): the minimum homology (similarity) between the query and the found sequence, as a fraction between 0 and 1. Only used if imperfect_homology is True. Defaults to 0.8.
        fixed_errors (bool, optional): if True, use a fixed number of errors instead of calculating it from min_homology. Only used if imperfect_homology is True. Defaults to False.
        inverted (bool, optional): if True, search for the reverse complement of the query sequence as well. Defaults to True.

    Returns:
        list: a list of query-match pairs, where each pair is a list containing the query sequence and a list of matching sequences.
    '''
    query_string=sequence[:query_length]

    if inverted == True:
        query = rev_compl(query_string)
    elif inverted == False:
        query = query_string
        
    sequence=sequence[query_length+min_spacer:]
    query_complement = rev_compl(query)
    output_pair_list=[]

    for i in range(len(sequence)-query_length+1):
        #print(sequence[i:query_length+i])
        if imperfect_homology:
            
            
            output_pair=imperfectHomologySearch(sequence[i:query_length+i],
                                                query,
                                                min_homology=min_homology,
                                                fixed_errors=fixed_errors,
                                                inverted=inverted)

            if output_pair != []:
                output_pair_list.append(output_pair)

        else:
            if sequence[i:query_length+i] == query:

                if inverted == True:
                    print(f"Success {query_complement}, {sequence[i:query_length+i]}")
                    output_pair_list.append([query_complement, [sequence[i:query_length+i]]])

                if inverted == False:
                    print(f"Success {query}, {sequence[i:query_length+i]}")
                    output_pair_list.append([query, [sequence[i:query_length+i]]])

    return(output_pair_list)
        
def searchSequenceForRepeats(sequence: str, min_query_length: int = 4, max_query_length: int = 25,
                            min_spacer: int = 0, window_size: int = 250, imperfect_homology:bool = False,
                            min_homology:float = 0.8, fixed_errors:bool = False, inverted: bool = True) -> dict:
    '''
    Search a DNA sequence for inverted repeats of various lengths.

    Args:
        sequence (str): the DNA sequence to search in.
        min_query_length (int, optional): the minimum length of the query sequence to search for. Defaults to 4.
        max_query_length (int, optional): the maximum length of the query sequence to search for. Defaults to 25.
        min_spacer (int, optional): the minimum number of non-matching nucleotides between the two halves of the inverted repeat. Defaults to 0.
        window_size (int, optional): the size of the sliding window to use when searching for inverted repeats. Defaults to 250.
        imperfect_homology (bool, optional): if True, allow for a certain number of errors in the inverted repeat. Defaults to False.
        min_homology (float, optional): the minimum homology (similarity) between the query and the found sequence, as a fraction between 0 and 1. Only used if imperfect_homology is True. Defaults to 0.8.
        fixed_errors (bool, optional): if True, use a fixed number of errors instead of calculating it from min_homology. Only used if imperfect_homology is True. Defaults to False.
        inverted (bool, optional): if True, search for the reverse complement of the query sequence as well. Defaults to True.

    Returns:
        dict: a dictionary where the keys are the query sequences and the values are lists of matching sequences.
    '''
    if imperfect_homology:
        print(f"Search has been set to find quasi-pallindromes")
        if fixed_errors:
            print(f"        Allowing up to {fixed_errors} errors/mismatches...")
        if not fixed_errors:
            print(f"        Searching with a minimum of {min_homology} homology")
    if not imperfect_homology:
        print(f"Search has been set to find perfect pallindromes")

    
    list_of_all_pairs=[]
    for query_length in range(min_query_length, max_query_length+1):
        print(f"###Searching for inverted-repeating {query_length}bp-long fragments...")
        sequence_size=len(sequence)
        
        for i in range(sequence_size): #-window_size+1
            seq=sequence[i:i+window_size]
            output_pair_list = findInvertedRepeat(  seq,
                                                    query_length=query_length,
                                                    min_spacer=min_spacer,
                                                    imperfect_homology=imperfect_homology,
                                                    min_homology=min_homology,
                                                    fixed_errors=fixed_errors,
                                                    inverted=inverted)
            for pair in output_pair_list:
                list_of_all_pairs.append(pair)

    results_dictionary={}
    print("Consolodating Results...")
    for pair in list_of_all_pairs:
        query=pair[0]
        if query in results_dictionary:
            if pair[1][0] not in results_dictionary[query]:
                results_dictionary[query].append(pair[1][0])
        else:
            results_dictionary[query] = [pair[1][0]]
    print("Results were consolidated...")
    return(results_dictionary)

def saveToJSON(results_dictionary: dict, output_file_name: str = 'json_output.json') -> None:
    '''
    Save a dictionary to a JSON file.

    Args:
        results_dictionary (dict): the dictionary to save.
        output_file_name (str, optional): the name of the output file. Defaults to 'json_output.json'.

    Returns:
        None
    '''
    import json
    with open(output_file_name, 'w') as outfile:
            json.dump(results_dictionary, outfile)




# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

## Contents:
## verifyImperfectHomology
## add_gene
## add_exon
## createAnnotatedOutput

import pandas as pd
from pyensembl import EnsemblRelease
from pyensembl.exon import Exon
import logging
import regex as re
from ..base import createChrList
from ..base import rev_compl
from ..complexity import movingWindow


def verifyImperfectHomology(ref, query, min_homology=0.8):

  mmbir = rev_compl(query)
  mmbir_errors = round(len(mmbir)*(1-min_homology))
  if mmbir_errors > 10:
    mmbir_errors = 10
  output_list = re.findall( '(' + mmbir + '){e<=' + str(mmbir_errors) + '}', ref)
  print(f'Found {len(output_list)} possible templates for {query}: {output_list}')

  if len(output_list) > 0:
    return(True)
  else:
    return(False)

def add_gene(row):

  data = EnsemblRelease(104)
  #print("loaded genome! #addgene")
  output = data.gene_names_at_locus(contig=row['chr'].strip("chr"), position=int(row["iBirStart"]))
  print(output)
  #output = str(output)
  out_list = []
  for gname in output:
    if gname != "":
      out_list.append(gname)
      #print("appended_gene!")

  return(out_list) #output

def add_exon(row):

  data = EnsemblRelease(104)
  #print("loaded genome! #addexon")
  output = data.exons_at_locus(contig=row['chr'].strip("chr"), position=int(row["iBirStart"]))
  #output = str(output)
  out_list = []
  for exon in output:
    if isinstance(exon, Exon):
      out_list.append(exon.exon_id)

  return(out_list) #output

def createAnnotatedOutput(f_name, path, output_f_name):
  import os

  chrList=createChrList(25)
  df = pd.DataFrame()

  for chr in chrList:
    try:
      if os.stat(f"{path}{chr}/{f_name}").st_size == 0:
        print(f"WARNING! File {path}{chr}/{f_name} is empty. Skipping...")
        continue
      with open(f"{path}{chr}/{f_name}") as f:
        lines = f.readlines()
    except:
      print(f"WARNING! Couldn't read {path}{chr}/{f_name}. Make sure it's there.")
      lines = None
      continue

    if (type(lines) == None):
      print("WARNING! the type of lines is None. Skipping...")

    else:
      print(f"hello {chr}")
      for i in range(len(lines)):
        if "###" in lines[i]:
          i-=1
          if "Microhomology Insertion:" in lines[i+11]:
            var0 = lines[i].strip()
            var1 = lines[i+1].strip()
            var2 = lines[i+2].strip("iBirStart:").strip()
            var3 = lines[i+3].strip("Consensus/cluster number:").strip()
            var4 = lines[i+4].strip("iDepth:").strip()
            var5 = lines[i+5].strip("sBir:").strip()
            var6 = lines[i+6].strip("sBirReversed:").strip()
            var7 = lines[i+7].strip()
            var8 = lines[i+8].strip("ref:").strip()
            var9 = lines[i+9].strip("bir:").strip()
            var10 = lines[i+10].strip("TEM:").strip()
            var11 = lines[i+11].strip("Microhomology Insertion:").strip()
            var12 = lines[i+12].strip("Microhomology template:").strip()
            var13 = lines[i+13].strip("readStart: ").strip()
            var14 = lines[i+14].strip("ref:").strip()
            var15 = lines[i+15].strip("bir:").strip()
            var16 = lines[i+16].strip()
            var17 = lines[i+17].strip()
            var18 = lines[i+18].strip()
            var19 = lines[i+19].strip()
          else:
            var0 = lines[i].strip()
            var1 = lines[i+1].strip()
            var2 = lines[i+2].strip("iBirStart:").strip()
            var3 = lines[i+3].strip("Consensus/cluster number:").strip()
            var4 = lines[i+4].strip("iDepth:").strip()
            var5 = lines[i+5].strip("sBir:").strip()
            var6 = lines[i+6].strip("sBirReversed:").strip()
            var7 = lines[i+7].strip()
            var8 = lines[i+8].strip("ref:").strip()
            var9 = lines[i+9].strip("bir:").strip()
            var10 = lines[i+10].strip("TEM:").strip()
            var11 = "Empty"
            var12 = "Empty"
            var13 = lines[i+11].strip("readStart: ").strip()
            var14 = lines[i+12].strip("ref:").strip()
            var15 = lines[i+13].strip("bir:").strip()
            var16 = lines[i+14].strip()
            var17 = lines[i+15].strip()
            var18 = lines[i+16].strip()

          event_dict = {
                        "chr":chr,
                        "iBirStart":var2,
                        "Consensus/cluster_number":var3,
                        "iDepth":var4,
                        "sBir":var5,
                        "sBirReversed":var6,
                        "ref":var8,
                        "bir":var9,
                        "TEM":var10,
                        "Microhomology_Insertion":var11,
                        "Microhomology_template":var12,
                        "readStart":var13,
                        "var16":var16,
                        "var17":var17,
                        "var18":var18
                        }
          
          #make sure that the ref and bir are not empty, otherwise the movingWindow will fail
          if (len(var8) == 0) or (len(var9) == 0):
            print(f"ref or bir are empty for {path}{chr}/{f_name}")
            print("Cheeck if they run correctly in the original output file...")
            print("Skipping this event...")

            #print the same messages to the error log
            logging.basicConfig(filename='error.log', level=logging.DEBUG)
            logging.debug(f"ref or bir are empty for {path}{chr}/{f_name}")
            logging.debug("Cheeck if they run correctly in the original output file...")
            logging.debug("Skipping this event...")
            continue

          event_df = pd.DataFrame.from_records([event_dict])
          df = pd.concat([df, event_df], ignore_index=True)

  print("Starting complexity")

  try:
    df[['ref_complexity_fail', 'ref_complexity_score', 'ref_complexity_tree']] = df.apply(lambda row: movingWindow(row.ref, complexity_threshold=0.2), axis=1, result_type ='expand')
  except:
    print(f"Error with ref_complexity for {path}{chr}/{f_name}")

  try:
    df[['bir_complexity_fail', 'bir_complexity_score', 'bir_complexity_tree']] = df.apply(lambda row: movingWindow(row.bir, complexity_threshold=0.2), axis=1, result_type ='expand')
  except:
    print(f"Error with bir_complexity for {path}{chr}/{f_name}")

  print("Starting homology check")
  df['homology_check_ref'] = df.apply(lambda row: verifyImperfectHomology(row.ref, row.sBir), axis=1)
  df['homology_check_bir'] = df.apply(lambda row: verifyImperfectHomology(row.bir, row.sBir), axis=1)
  print("Starting gene/exon annotation")
  df["genes"] = df.apply(lambda row: add_gene(row), axis=1)
  df["exones"] = df.apply(lambda row: add_exon(row), axis=1)

  df.to_csv(output_f_name, sep="\t",index=False)
  print(f"finished annotation... DF saved to {output_f_name}")


from setuptools import setup, find_packages
import codecs
import os


VERSION = '0.3.0'
DESCRIPTION = 'Genetic Analysis Tools'
LONG_DESCRIPTION = 'A package that contains tools for the analysis of genetic sequence and sequencing related data'

# Setting up
setup(
    name="bio-aid",
    version=VERSION,
    author="tvarovski (Jerzy Twarowski)",
    author_email="<tvarovski1@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['pandas', 'numpy', 'matplotlib', 'seaborn', 'regex', 'pyensembl', 'natsort'],
    keywords=['python', 'biology', 'bio', 'genetics', 'genomics', 'NGS'],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.10",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

def calculate_dilution(count_per_square: int|float, final_plate: int|float) -> None:
    """
    Calculates the dilution factor and dilution amounts for a given cell count and final plate volume.

    Args:
        count_per_square (float): The number of cells counted per square on a hemocytometer.
        final_plate (float): The desired number of cells per plate.

    Returns:
        None

    Examples:
        >>> calculate_dilution(50, 100)
        
        You should dilute by using 0.25 of cell culture with 0.75uL ddH2O 
        and then with three steps of 100uL of diluted culture and 900uL ddH2O.
        Plate with 100uL of diluted solution per plate
    """
    init_cells_ml = (count_per_square/4.0)*10**6
    final_cells_ml = final_plate*10

    dilution_factor = init_cells_ml/final_cells_ml
    dilution_factor = dilution_factor/1000

    fourth_dilution_amount = 1000/dilution_factor

    print(f"\nYou should dilute by using {fourth_dilution_amount} of cell culture with {1000-fourth_dilution_amount}uL ddH2O \nand then with three steps of 100uL of diluted culture and 900uL ddH2O.\nPlate with 100uL of diluted solution per plate")

def dilute() -> None:
    """
    Prompts the user for input and calculates the dilution factor and dilution amounts using the calculate_dilution function.

    Args:
        None

    Returns:
        None

    Raises:
        None

    Examples:
        >>> dilute()

        Welcome to the Diluter 1.0. Enjoy not doing the math yourself!

        Initial number of cells refers to the average number of cells on the 1/16th of hematocrit slide.
        For best results take the average of at least three of these squares.

        Type initial number of cells in media or enter 'x' to exit: 50

        Type in desired number of colonies per plate (100uL of media): 100

        You should dilute by using 0.25 of cell culture with 0.75uL ddH2O 
        and then with three steps of 100uL of diluted culture and 900uL ddH2O.
        Plate with 100uL of diluted solution per plate
    """

    welcome_message = "\nWelcome to the Diluter 1.0. Enjoy not doing the math yourself!\n"
    print(welcome_message)

    print("Initial number of cells refers to the average number of cells on the 1/16th of hematocrit slide.\nFor best results take the average of at least three of these squares.")

    looping = True

    while looping:
        initial_cells = input("\nType initial number of cells in media or enter 'x' to exit: ")
        if initial_cells == "x":
            looping = False
            continue
        initial_cells = float(initial_cells)
        final_cells = input("\nType in desired number of colonies per plate (100uL of media): ")
        final_cells = float(final_cells)

        calculate_dilution(initial_cells,final_cells)


# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

import regex as re

def createMMBSearchReference(file_path_in, file_path_out):
  #good for human genome
  file_in = open(file_path_in, "r")
  save=False
  readlines=True
  while readlines:
    try:
      line = file_in.readline()
    except:
      print("Failed to read line. EOF? Exiting...")
      readlines=False
      break

    if line[0] == ">":
      if re.search("chromosome.*Primary.Assembly$", line) != None:
        print(line)
        linewords = line.split()
        chromosome=linewords[4].strip(",")
        if (chromosome=="1") | (chromosome=="2"):
          chromosome=f">chr0{chromosome}\n"
        elif chromosome=="X":
          chromosome=f">chrX\n"
        elif chromosome=="Y":
          chromosome=f">chrY\n"
        else:
          chromosome=f">chr{chromosome}\n"
        line=chromosome
        print(line)
        save=True
      #fix for mitochondrial chr
      elif re.search("mitochondrion, complete genome$", line) != None:
        print(line)
        line=f">chrM\n"
        print(line)
        save=True
      else:
        save=False

    if save:
      with open(file_path_out, "a") as file_out:
        file_out.write(line)


# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

## Contents:
## wordsInSequence
## findComplexity
## movingWindow
## createLogFile
## createSequenceList
## createDataFrameFromLogFile

import pandas as pd
import ast
from pandas import DataFrame
from .base import unique

def wordsInSequence(sequence: str, treeLevel: int) -> list:
    """
    Extracts all possible words of a specified length from a genetic sequence.

    Args:
        sequence (str): The genetic sequence to extract words from.
        treeLevel (int): The length of the words to extract.

    Returns:
        list: A list of words, where each word is a string of length treeLevel.

    Examples:
        >>> wordsInSequence("ATCGATCG", 3)
        ['ATC', 'TCG', 'CGA', 'GAT', 'ATC', 'TCG']

        >>> wordsInSequence("ATCGATCG", 4)
        ['ATCG', 'TCGA', 'CGAT', 'GATC', 'ATCG']
    """
    length = len(sequence)
    max_possible_words_in_sequence = length-(treeLevel-1)
    wordList = []
    for i in range(max_possible_words_in_sequence):
        wordList.append(sequence[0+i:treeLevel+i])
    return wordList

def findComplexity(sequence, treeLevel, complexity_threshold):
    # takes a genetic sequence and calculates the linguistic complexity 
    # for that sequence for word length treeLevel

    wordList = wordsInSequence(sequence,treeLevel)
    wordList = unique(wordList)

    if len(sequence) > 4**treeLevel:
        complexity = len(wordList)/(4**treeLevel)
    else:
        complexity = len(wordList)/(len(sequence)-(treeLevel-1))

    if (complexity < complexity_threshold) & (complexity > 0.2):
        print("Complexity at tree level " + str(treeLevel) + " is " + str(complexity) + " for sequence: "+str(sequence))

    return ([complexity, treeLevel, sequence])

def movingWindow(sequence: str, treeLevel: int = 8, chunkSize: int = 20, complexity_threshold: float = 0.2, showGraph: bool = False) -> list:
    """
    Calculates the linguistic complexity scores for windows of a specified size and word lengths from 1 to a specified level.

    Args:
        sequence (str): The genetic sequence to calculate complexity scores for.
        treeLevel (int, optional): The maximum length of words to use in the complexity calculation. Defaults to 8.
        chunkSize (int, optional): The size of the windows to use in the complexity calculation. Defaults to 20.
        complexity_threshold (float, optional): The threshold below which a window is considered to have low complexity. Defaults to 0.2.
        showGraph (bool, optional): Whether to show a graph of the complexity scores. Defaults to False.

    Returns:
        list: A list containing the lowest complexity score window in the format [Boolean, complexity, treeLevel], where Boolean denotes if the complexity score is lower (True) than complexity_threshold.

    Raises:
        ValueError: If the sequence is too short to calculate complexity scores.

    Examples:
        >>> movingWindow("ATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCG", treeLevel=4, chunkSize=10, complexity_threshold=0.1, showGraph=True)
        [True, 0.0, 4]

        >>> movingWindow("ATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCGATCG", treeLevel=10, chunkSize=5, complexity_threshold=0.5, showGraph=True)
        [False, 0.5, 10]
    """

    complexityList=[]

    for i in range(1, treeLevel+1):
        for chunk in range(len(sequence)-(chunkSize-1)):

            chunkSeq = sequence[chunk:(chunk+chunkSize)]
            complexityList.append(findComplexity(chunkSeq, i, complexity_threshold))

    try:
        lowest=min(complexityList, key=lambda x: x[0])

    except ValueError:
        print(f"ERROR with sequence: {sequence}. It is probably too short len={len(sequence)}. Skipping it.")
        print("complexityList: ", complexityList)
        lowest = [0,0,0]

    #print(complexityList)

    lowLvl=lowest[1]

    if showGraph:

        df = DataFrame(complexityList,columns=['complexity','tree','sequence'])
        df = df[['complexity','tree']]
        ax = df[df.tree == lowLvl].filter(items=["complexity"]).plot(fontsize=15, grid=True, figsize=(20,8))
        ax.axhline(complexity_threshold, color="red", linestyle="--")
        ax.axhline(1, color="green", linestyle="--")
        ax.set_ylim(ymin=0)
        ax.set_title(str(lowest[2]))

    if lowest[0] < complexity_threshold:
        return([True, lowest[0], lowest[1]])
        #return([True, lowest])
    else:
        return([False, lowest[0], lowest[1]])
        #return([False, lowest])

def createLogFile(seqlist: list, filename: str, complexity_threshold: float = 0.2, chunkSize: int = 20) -> None:
    """
    Creates a log file containing the lowest linguistic complexity score for each sequence in a list.

    Args:
        seqlist (list): A list of genetic sequences to calculate complexity scores for.
        filename (str): The name of the output file to write the log to.
        complexity_threshold (float, optional): The threshold below which a sequence is considered to have low complexity. Defaults to 0.2.
        chunkSize (int, optional): The size of the windows to use in the complexity calculation. Defaults to 20.

    Returns:
        None
    """
    outputf = open(filename, "w")
    truemmbir = 0
    for i in seqlist:
        seqScore = movingWindow(i, complexity_threshold = complexity_threshold, chunkSize = chunkSize)
        if seqScore[0] == False:
            truemmbir += 1
        outputf.write(str(seqScore[1])+"\n")
    print("sequences above threshold:", truemmbir)

    outputf.close()

def createSequenceList(filename: str) -> list:
    """
    Extracts genetic sequences from a file and returns them as a list.

    Args:
        filename (str): The name of the file to extract sequences from.

    Returns:
        list: A list of genetic sequences, where each sequence is a string.
    """
    seqfile = open(filename, "r")
    lines = seqfile.readlines()
    seqlist=[]
    for line in lines:
        if "bir:" in line:
            if len(line) > 100:
                seqlist.append(line.strip("\n").strip("bir:-"))
    seqlist = list(dict.fromkeys(seqlist))
    print(len(seqlist))

    longseq=0
    for i in seqlist:
        if len(i) > 300:
            longseq+=1
    print(longseq)

    return seqlist

def createDataFrameFromLogFile(logfile_path: str) -> DataFrame:
    """
    Creates a pandas DataFrame from a log file containing linguistic complexity scores.

    Args:
        logfile (str): The name of the log file to create the DataFrame from.

    Returns:
        DataFrame: A pandas DataFrame containing the complexity scores, where each row represents a sequence and contains the complexity score, the word length, and the sequence itself.
    """
    logfile = open(logfile_path, "r")
    lines = logfile.readlines()
    loglist=[]
    for line in lines:
        if "[" in line:
                loglist.append(line.strip("\n"))

    for i in range(len(loglist)):
        loglist[i] = ast.literal_eval(loglist[i]) 
    print("total sequences:", len(loglist))

    df = DataFrame(loglist,columns=['complexity', 'level', 'sequence'])
    
    return df


# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

from .AnnoMMBS import createAnnotatedOutput
from .makeMMBSearchRef import createMMBSearchReference


# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

## Contents:
## split_df
## join_slices
## clean_up_slices

import pandas as pd
import logging

def split_df(df: pd.DataFrame, slices_num: int, savecsv_path_root=None) -> list:
    """
    Splits a pandas DataFrame into n=slices_num slices and saves each slice to a CSV file if a path is provided.
    Returns a list of the DataFrame slices.

    Args:
        df (pandas.DataFrame): The DataFrame to be split.
        slices_num (int): The number of slices to create.
        savecsv_path_root (str): Optional. The root path to save the CSV files. If not provided, the slices are not saved.

    Returns:
        list: A list of the DataFrame slices.
    """

    from math import ceil

    slice_size = ceil(len(df) / slices_num)
    slices_out = []
    for slice in range(slices_num):
        df_slice = df[slice*slice_size:(slice+1)*slice_size].copy()
        
        if savecsv_path_root:
            df_slice_name = f"{savecsv_path_root[0]}_{slice}.csv"
            df_slice.to_csv(df_slice_name, index=False)
            logging.info(f'Slice {slice} saved to {df_slice_name}')
        
        slices_out.append(df_slice)
    
    logging.info(f'Input df split into {slices_num} slices.')
    return slices_out

def join_slices(slices_num: int, slice_path_root: str, savecsv_path=None):
    """
    Combines the results from each slice into one pandas DataFrame and saves it to a CSV file if a path is provided.
    Returns the combined DataFrame.

    Args:
        slices_num (int): The number of slices to combine.
        slice_path_root (str): The root path to the CSV files for each slice.
        *savecsv_path (str): Optional. The path to save the combined DataFrame as a CSV file.

    Returns:
        pandas.DataFrame: The combined DataFrame.
    """

    df = pd.DataFrame()
    for slice in range(slices_num):
        df_slice = pd.read_csv(f"{slice_path_root}_{slice}.csv")
        df = pd.concat([df, df_slice], ignore_index=True)
    
    if savecsv_path:
        df.to_csv(savecsv_path[0], index=False)
        logging.info(f'Slices joined! Results saved to {savecsv_path[0]}')
    
    return df

def clean_up_slices(slices_num: int, slice_path_root: str) -> None:
    """
    Deletes the CSV files for each slice of a DataFrame that was split using the `split_df` function.

    Args:
        slices_num (int): The number of slices that were created.
        slice_path_root (str): The root path to the CSV files for each slice.

    Returns:
        None
    """
    import os
    for slice in range(slices_num):
        os.remove(f"{slice_path_root}_{slice}.csv")

    logging.info(f'Clean up Finished! Deleted slice files.')


# This code was developed and authored by Jerzy Twarowski in Malkova Lab at the University of Iowa 
# Contact: jerzymateusz-twarowski@uiowa.edu, tvarovski1@gmail.com

import regex as re

def createMMBSearchReference(file_path_in: str, file_path_out: str) -> None:
    """
    Reads a FASTA file containing genome data and creates a new file with a modified header line for each chromosome.
    It's purpose is to create a FASTA file that can be used as a reference for MMBSearch.

    Args:
        file_path_in (str): The path to the input FASTA file.
        file_path_out (str): The path to the output file to be created.

    Returns:
        None
    """
    #good for human genome
    file_in = open(file_path_in, "r")
    save=False
    readlines=True
    while readlines:
        try:
            line = file_in.readline()
        except:
            print("Failed to read line. EOF? Exiting...")
            readlines=False
            break

        if line[0] == ">":
            if re.search("chromosome.*Primary.Assembly$", line) != None:
                print(line)
                linewords = line.split()
                chromosome=linewords[4].strip(",")
                if (chromosome=="1") | (chromosome=="2"):
                    chromosome=f">chr0{chromosome}\n"
                elif chromosome=="X":
                    chromosome=f">chrX\n"
                elif chromosome=="Y":
                    chromosome=f">chrY\n"
                else:
                    chromosome=f">chr{chromosome}\n"
                line=chromosome
                print(line)
                save=True
            #fix for mitochondrial chr
            elif re.search("mitochondrion, complete genome$", line) != None:
                print(line)
                line=f">chrM\n"
                print(line)
                save=True
            else:
                save=False

        if save:
            with open(file_path_out, "a") as file_out:
                file_out.write(line)